from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import textwrap

import pytest

try:
    EXIT_OK = pytest.ExitCode.OK
    EXIT_TESTSFAILED = pytest.ExitCode.TESTS_FAILED
    EXIT_USAGEERROR = pytest.ExitCode.USAGE_ERROR
except AttributeError:
    # ExitCode was introduced in pytest 5.0.0. As long as we support pytest 4.6 (for
    # Python 2), use the private constants for readability.
    from _pytest.main import EXIT_OK, EXIT_TESTSFAILED, EXIT_USAGEERROR  # noqa: F401

from pytest_quarantine import DEFAULT_QUARANTINE


def test_default_file():
    assert DEFAULT_QUARANTINE == "quarantine.txt"


def test_options(testdir):
    result = testdir.runpytest("--help")

    result.stdout.fnmatch_lines(
        ["quarantine:", "*--save-quarantine=*", "*--quarantine=*"]
    )


@pytest.fixture
def error_failed_passed(testdir):
    """Create test_error_failed_passed.py with one test for each outcome."""
    return testdir.makepyfile(
        test_error_failed_passed="""\
        import pytest

        @pytest.fixture
        def error():
            assert False

        def test_error(error):
            assert True

        def test_failed():
            assert False

        def test_passed():
            assert True
        """
    )


def test_no_output_without_options(testdir):
    result = testdir.runpytest()
    assert DEFAULT_QUARANTINE not in result.stdout.str()


@pytest.fixture
def testdir(testdir):
    """Return default testdir fixture with additional helper methods."""

    def _path_has_content(path, content):
        return testdir.tmpdir.join(path).read() == textwrap.dedent(content)

    testdir.path_has_content = _path_has_content

    def _write_path(path, content):
        testdir.tmpdir.join(path).write(textwrap.dedent(content))

    testdir.write_path = _write_path

    return testdir


@pytest.mark.parametrize("quarantine_path", [DEFAULT_QUARANTINE, ".quarantine"])
def test_save_failing_tests(quarantine_path, testdir, error_failed_passed):
    args = ["--save-quarantine"]
    if quarantine_path != DEFAULT_QUARANTINE:
        args.append(quarantine_path)

    result = testdir.runpytest(*args)

    result.stdout.fnmatch_lines(
        ["*- 2 items saved to {} -*".format(quarantine_path), "=*failed*"]
    )
    result.assert_outcomes(passed=1, failed=1, error=1)
    assert result.ret == EXIT_TESTSFAILED

    assert testdir.path_has_content(
        quarantine_path,
        """\
        test_error_failed_passed.py::test_error
        test_error_failed_passed.py::test_failed
        """,
    )


def test_dont_save_other_outcomes(testdir):
    quarantine_path = DEFAULT_QUARANTINE
    args = ["--save-quarantine"]

    testdir.makepyfile(
        """\
        import pytest

        def test_passed():
            assert True

        @pytest.mark.skip
        def test_skipped():
            assert False

        @pytest.mark.xfail
        def test_xfailed():
            assert False

        @pytest.mark.xfail
        def test_xpassed():
            assert True
        """
    )

    result = testdir.runpytest(*args)

    result.stdout.fnmatch_lines(
        ["*- 0 items saved to {} -*".format(quarantine_path), "=*skipped*"]
    )
    result.assert_outcomes(passed=1, skipped=1, xpassed=1, xfailed=1)
    assert result.ret == EXIT_OK

    assert testdir.path_has_content(quarantine_path, "")


def test_save_empty_quarantine(testdir):
    quarantine_path = DEFAULT_QUARANTINE
    args = ["--save-quarantine"]

    testdir.makepyfile(
        test_xpassed="""\
        def test_passed():
            assert True
        """
    )

    testdir.write_path(
        quarantine_path,
        """\
        test_xpassed.py::test_passed
        """,
    )

    result = testdir.runpytest(*args)

    result.stdout.fnmatch_lines(
        ["*- 0 items saved to {} -*".format(quarantine_path), "=*passed*"]
    )
    result.assert_outcomes(passed=1)
    assert result.ret == EXIT_OK

    assert testdir.path_has_content(quarantine_path, "")


@pytest.mark.parametrize("quarantine_path", [DEFAULT_QUARANTINE, ".quarantine"])
def test_missing_quarantine(quarantine_path, testdir):
    args = ["--quarantine"]
    if quarantine_path != DEFAULT_QUARANTINE:
        args.append(quarantine_path)

    result = testdir.runpytest(*args)

    result.stderr.fnmatch_lines(
        ["ERROR: Could not load quarantine:*'{}'".format(quarantine_path)]
    )
    assert result.ret == EXIT_USAGEERROR


@pytest.mark.parametrize("quarantine_path", [DEFAULT_QUARANTINE, ".quarantine"])
def test_full_quarantine(quarantine_path, testdir, error_failed_passed):
    args = ["--quarantine"]
    if quarantine_path != DEFAULT_QUARANTINE:
        args.append(quarantine_path)

    testdir.write_path(
        quarantine_path,
        """\
        test_error_failed_passed.py::test_error
        test_error_failed_passed.py::test_failed
        """,
    )

    result = testdir.runpytest(*args)

    result.stdout.fnmatch_lines(
        [
            "collected*",
            "added mark.xfail to 2 of 2 items from {}".format(quarantine_path),
        ]
    )
    result.assert_outcomes(passed=1, xfailed=2)
    assert result.ret == EXIT_OK


def test_partial_quarantine(testdir, error_failed_passed):
    quarantine_path = DEFAULT_QUARANTINE
    args = ["--quarantine"]

    testdir.write_path(
        quarantine_path,
        """\
        test_error_failed_passed.py::test_failed
        test_error_failed_passed.py::test_extra
        """,
    )

    result = testdir.runpytest(*args)

    result.stdout.fnmatch_lines(
        [
            "collected*",
            "added mark.xfail to 1 of 2 items from {}".format(quarantine_path),
        ]
    )
    result.assert_outcomes(passed=1, error=1, xfailed=1)
    assert result.ret == EXIT_TESTSFAILED


def test_only_extra_quarantine(testdir, error_failed_passed):
    quarantine_path = DEFAULT_QUARANTINE
    args = ["--quarantine"]

    testdir.write_path(
        quarantine_path,
        """\
        test_error_failed_passed.py::test_extra
        """,
    )

    result = testdir.runpytest(*args)

    result.stdout.fnmatch_lines(
        [
            "collected*",
            "added mark.xfail to 0 of 1 item from {}".format(quarantine_path),
        ]
    )
    result.assert_outcomes(passed=1, failed=1, error=1)
    assert result.ret == EXIT_TESTSFAILED


def test_passing_quarantine(testdir, error_failed_passed):
    quarantine_path = DEFAULT_QUARANTINE
    args = ["--quarantine"]

    testdir.write_path(
        quarantine_path,
        """\
        test_error_failed_passed.py::test_error
        test_error_failed_passed.py::test_failed
        test_error_failed_passed.py::test_passed
        """,
    )

    result = testdir.runpytest(*args)

    result.stdout.fnmatch_lines(
        [
            "collected*",
            "added mark.xfail to 3 of 3 items from {}".format(quarantine_path),
        ]
    )
    result.assert_outcomes(xfailed=2, xpassed=1)
    assert result.ret == EXIT_OK


def test_no_report_with_quiet_option(testdir, error_failed_passed):
    quarantine_path = DEFAULT_QUARANTINE
    args = ["-q", "--quarantine"]

    testdir.write_path(
        quarantine_path,
        """\
        test_error_failed_passed.py::test_error
        """,
    )

    result = testdir.runpytest(*args)

    assert quarantine_path not in result.stdout.str()



"""A plugin for pytest to manage expected test failures.

Saves the list of failing tests, so that they can be automatically marked as expected
failures on future test runs.

Inspired by:
    - https://github.com/pytest-dev/pytest/blob/master/src/_pytest/cacheprovider.py
    - https://github.com/hackebrot/pytest-md/blob/master/src/pytest_md/plugin.py
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import attr
import pytest


# TODO: Guarantee this is opened from pytest's root dir
# (to allow running pytest in a subdirectory)
DEFAULT_QUARANTINE = "quarantine.txt"


def _item_count(nodeids):
    count = len(nodeids)
    return "{} item{}".format(count, "" if count == 1 else "s")


@attr.s(cmp=False)
class SaveQuarantinePlugin(object):
    """Save the list of failing tests to a quarantine file.

    Attributes:
        quarantine_path (str): The path to the quarantine file
        quarantine_ids (Set[int]):
            The ID's of failed tests to write to ``quarantine_path``
    """

    quarantine_path = attr.ib()
    quarantine_ids = attr.ib(init=False, factory=set)

    def pytest_runtest_logreport(self, report):
        """Save the ID of a failed test to the quarantine."""
        if report.failed:
            self.quarantine_ids.add(report.nodeid)

    def pytest_terminal_summary(self, terminalreporter):
        """Display size of quarantine after running tests."""
        terminalreporter.write_sep(
            "-",
            "{} saved to {}".format(
                _item_count(self.quarantine_ids), self.quarantine_path
            ),
        )

    def pytest_sessionfinish(self, session):
        """Write the ID's of quarantined tests to a file.

        Writes an empty file if all tests are passing, to remove previously failed tests
        from the quarantine.
        """
        with open(self.quarantine_path, "w") as f:
            f.writelines(nodeid + "\n" for nodeid in sorted(self.quarantine_ids))


@attr.s(cmp=False)
class QuarantinePlugin(object):
    """Mark each test listed in a quarantine file as xfail.

    Attributes:
        quarantine_path (str): The path to the quarantine file
        verbose (int): The value of pytest's "verbose" config option
        quarantine_ids (Set[int]): The test ID's read from ``quarantine_path``
        marked_ids (Set[int]): The ID's of collected tests marked as xfail
    """

    quarantine_path = attr.ib()
    verbose = attr.ib()
    quarantine_ids = attr.ib(init=False, factory=set)
    marked_ids = attr.ib(init=False, factory=set)

    def pytest_sessionstart(self, session):
        """Read test ID's from a file into the quarantine."""
        try:
            with open(self.quarantine_path) as f:
                self.quarantine_ids = {nodeid.strip() for nodeid in f}
        except IOError as exc:
            raise pytest.UsageError("Could not load quarantine: " + str(exc))

    def pytest_itemcollected(self, item):
        """Mark a test as xfail if its ID is in the quarantine."""
        if item.nodeid in self.quarantine_ids:
            item.add_marker(pytest.mark.xfail(reason="Quarantined"))
            self.marked_ids.add(item.nodeid)

    def pytest_report_collectionfinish(self):
        """Display number of quarantined items before running tests."""
        if self.verbose >= 0:
            return "added mark.xfail to {} of {} from {}".format(
                len(self.marked_ids),
                _item_count(self.quarantine_ids),
                self.quarantine_path,
            )


def pytest_configure(config):
    """Register the plugin functionality."""
    save_quarantine_path = config.getoption("save_quarantine")
    if save_quarantine_path:
        config.pluginmanager.register(
            SaveQuarantinePlugin(save_quarantine_path), "save_quarantine_plugin"
        )

    quarantine_path = config.getoption("quarantine")
    if quarantine_path:
        config.pluginmanager.register(
            QuarantinePlugin(quarantine_path, config.getoption("verbose")),
            "quarantine_plugin",
        )


def pytest_addoption(parser):
    """Add command line options to the 'quarantine' group."""
    group = parser.getgroup("quarantine")

    group.addoption(
        "--save-quarantine",
        nargs="?",
        const=DEFAULT_QUARANTINE,
        metavar="PATH",
        help="Write failing tests to %(metavar)s (default: %(const)s)",
    )

    group.addoption(
        "--quarantine",
        nargs="?",
        const=DEFAULT_QUARANTINE,
        metavar="PATH",
        help="Mark tests listed in %(metavar)s with `xfail` (default: %(const)s)",
    )



import sys

from setuptools_scm import get_version


def main():
    """Validate the package version for PyPI.

    If the version contains a local identifier, return an error message suitable for
    sys.exit. Otherwise, return None.

    See also:
    https://github.com/pypa/setuptools_scm/issues/365
    https://github.com/pypa/twine/issues/430
    https://www.python.org/dev/peps/pep-0440/#local-version-identifiers

    """
    version = get_version(root="..", relative_to=__file__)
    print("Version: " + version)

    if "+" in version:
        return (
            "ERROR: Invalid version. "
            "Make sure the working tree is clean, then tag a new version."
        )


if __name__ == "__main__":
    sys.exit(main())



import sys
from pathlib import Path
from shutil import rmtree


def main():
    """Remove files created by setup.py."""
    root = Path(__file__).parents[1].resolve()

    for dirname in ["build", "dist"]:
        rmtree(root / dirname, ignore_errors=True)


if __name__ == "__main__":
    sys.exit(main())
